import { stringify } from 'node:querystring'
import { is, has, reject, isNil, mergeDeepWith, concat, difference } from 'ramda'
import { IndexDescription, IndexDescriptionInfo, IndexDirection, ObjectId } from 'mongodb'
import { DEFAULT_CONNECTION_NAME, ENTITY_METADATA, ENTITY_WATERMARK, NESTED_ENTITY_WATERMARK } from './constants'
import { Metadata } from './metadata'
import {
    Class,
    DataSourceOptions,
    DataSourceSingleHost,
    EntityColumn,
    EntityColumnTypes,
    EntityColumnValidator,
    EntityIndex,
    EntityMetadata,
    MixedList,
    MongoJsonSchema
} from './types'
import { BSONType } from './enums'

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export const isDataSourceOptions = (object: any): object is DataSourceOptions => is(Object, object) && has('database', object)

// eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-argument
export const isEmpty = (value: any) => (Array.isArray(value) ? value.length === 0 : Object.keys(value).length === 0)

export const getClientToken = (connectionName?: string) => `mongo-client:${connectionName ?? DEFAULT_CONNECTION_NAME}`

export const getDataSourceToken = (connectionName?: string) => `mongo-datasource:${connectionName ?? DEFAULT_CONNECTION_NAME}`

export const getCollectionToken = (entity: Class, connectionName?: string) =>
    `mongo-entity:${connectionName ?? DEFAULT_CONNECTION_NAME}:${entity.name}`

export const getEntityMetadata = <T>(entity: Class<T>) => Metadata.get<EntityMetadata>(entity, ENTITY_METADATA)

export const getConnectionString = (options: DataSourceOptions): string => {
    const auth = options.username && options.password ? `${options.username}:${options.password}@` : ''
    const database = options.database ?? ''
    const hosts: Array<DataSourceSingleHost> = options.hosts ?? [
        {
            host: 'localhost',
            port: 27017
        }
    ]

    const extraOptions = stringify(
        reject(isNil, {
            authSource: options.authSource ?? 'admin',
            replicaSet: options.replicaSet
        })
    )

    const nodes = hosts.map(({ host, port }) => `${host}:${port}`).join(',')

    return `mongodb://${auth}${nodes}/${database}?${extraOptions}`
}

export const getDefaultDataSourceOptions = (options: DataSourceOptions): DataSourceOptions => ({
    ...options
})

export const getEntitiesFromMixedList = (mixedList: MixedList<Class> = []): Array<Class> => {
    const entities = Array.isArray(mixedList) ? mixedList : Object.values(mixedList ?? {})

    return entities.reduce((acc, entity) => (!Metadata.has(entity, ENTITY_WATERMARK) ? acc : [...acc, entity]), [] as Array<Class>)
}

export const getBsonType = (type: EntityColumnTypes) => {
    switch (type) {
        case String:
            return BSONType.String
        case Number:
            return BSONType.Number
        case Boolean:
            return BSONType.Bool
        case Date:
            return BSONType.Date
        case Array:
            return BSONType.Array
        case ObjectId:
            return BSONType.ObjectId
        default:
            return BSONType.Object
    }
}

export const getPropertyValidator = (column: EntityColumn): EntityColumnValidator => {
    const bsonType = getBsonType(column.type)
    const isNestedEntity = Metadata.has(column.type, NESTED_ENTITY_WATERMARK)

    if (isNestedEntity && column.options.array) {
        const metadata = Metadata.get<EntityMetadata>(column.type, ENTITY_METADATA)

        return {
            bsonType: BSONType.Array,
            additionalProperties: false,
            items: getObjectValidator(metadata.entityColumns)
        }
    }

    if (isNestedEntity) {
        const metadata = Metadata.get<EntityMetadata>(column.type, ENTITY_METADATA)

        return getObjectValidator(metadata.entityColumns)
    }

    if (column.options.array && column.options.enum) {
        return {
            bsonType: BSONType.Array,
            additionalProperties: false,
            items: {
                enum: Object.values(column.type)
            }
        }
    }

    if (column.options.enum) {
        return reject(isEmpty, {
            enum: Object.values(column.options.enum)
        })
    }

    if (column.options.array) {
        return {
            bsonType: BSONType.Array,
            additionalProperties: false,
            items: {
                bsonType
            }
        }
    }

    return reject(isNil, {
        bsonType
    })
}

export const getObjectValidator = (columns: Array<EntityColumn>) =>
    reject(
        isEmpty,
        columns.reduce<EntityColumnValidator>(
            (acc, column) => ({
                bsonType: BSONType.Object,
                additionalProperties: false,
                required: column.options.required ? [...(acc.required ?? []), column.propertyKey] : acc.required,
                properties: {
                    ...acc.properties,
                    [column.propertyKey]: getPropertyValidator(column)
                }
            }),
            {}
        )
    )

export const getEntitySchemaValidator = (columns: Array<EntityColumn>): MongoJsonSchema => ({
    $jsonSchema: getObjectValidator(columns)
})

export const getIndexNameFromKey = (key: string) => key.replace(/\.+/g, '_').toUpperCase()

// eslint-disable-next-line @typescript-eslint/no-explicit-any
export const mergeObjectAndArray = (left: any, right: any): any => {
    if (Array.isArray(left) && Array.isArray(right)) {
        return concat(left, right)
    }

    if (is(Object, left) && is(Object, right)) {
        return mergeObjectAndArray(left, right)
    }

    return right
}

export const getEntityIndexes = (metadata: EntityMetadata, stackedKeys: Array<string> = []): Record<string, EntityIndex> => {
    const columns = metadata.entityColumns
    const createIndexOptionMap = metadata.entityOptions.index ?? {}

    return columns.reduce(
        (result, column) => {
            const isNestedEntity = Metadata.has(column.type, NESTED_ENTITY_WATERMARK)

            if (isNestedEntity) {
                const nestedEntityMetadata = Metadata.get<EntityMetadata>(column.type, ENTITY_METADATA)

                return {
                    ...result,
                    ...getEntityIndexes(nestedEntityMetadata, [...stackedKeys, column.propertyKey])
                }
            }

            if (isNil(column.options.index)) {
                return result
            }

            const key = [...stackedKeys, column.propertyKey].join('.')

            if (is(String, column.options.index)) {
                const { index: indexName } = column.options

                if (!createIndexOptionMap[indexName]) {
                    throw new Error(
                        `The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`
                    )
                }

                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return mergeDeepWith(mergeObjectAndArray, result, {
                    [indexName]: {
                        keys: [key],
                        options: createIndexOptionMap[indexName]
                    }
                })
            }

            if (Array.isArray(column.options.index)) {
                const { index: indexNames } = column.options

                indexNames.forEach(indexName => {
                    if (!createIndexOptionMap[indexName]) {
                        throw new Error(
                            `The index '${indexName}' on property '${column.propertyKey}' in [class ${column.parent.name}] has no index defined. Please add 'index[${indexName}]' in @Entity() or @NestedEntity() options.`
                        )
                    }
                })

                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return mergeDeepWith(
                    mergeObjectAndArray,
                    result,
                    indexNames.reduce(
                        (acc, indexName) =>
                            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                            mergeDeepWith(mergeObjectAndArray, acc, {
                                [indexName]: {
                                    keys: [key],
                                    options: createIndexOptionMap[indexName]
                                }
                            }),
                        {} as Record<string, EntityIndex>
                    )
                )
            }

            const indexName = getIndexNameFromKey(key)
            const { index: createIndexOptions } = column.options

            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            return mergeDeepWith(mergeObjectAndArray, result, {
                [indexName]: {
                    keys: [key],
                    options: createIndexOptions
                }
            })
        },
        {} as Record<string, EntityIndex>
    )
}

export const getEntityIndexDescriptions = (entityIndexes: Record<string, EntityIndex>): Array<IndexDescription> =>
    Object.entries(entityIndexes).map(([name, { keys, options }]) => ({
        name,
        key: keys.reduce(
            (acc, key) => ({
                ...acc,
                [key]: 1 // todo: add support for key direction to be specified on @Column() generator or @Entity() or @NestedEntity()
            }),
            {} as Record<string, IndexDirection>
        ),
        ...options
    }))

export const getEntityDropIndexNames = (currentIndexDescriptions: Array<IndexDescriptionInfo>, entityIndexes: Record<string, EntityIndex>) =>
    currentIndexDescriptions.reduce((acc, { name, key: currentIndexKeys, ...currentIndexOptions }) => {
        if (!name || name === '_id_') {
            return acc
        }

        const targetEntityIndex = entityIndexes[name]

        if (!targetEntityIndex) {
            return [...acc, name]
        }

        const differenceInKeys = difference(Object.keys(currentIndexKeys), targetEntityIndex.keys)

        if (differenceInKeys.length) {
            return [...acc, name]
        }

        const [differenceInOptions] = difference([currentIndexOptions], [targetEntityIndex.options])

        if (differenceInOptions && Object.keys(differenceInOptions).some(key => key !== 'v')) {
            return [...acc, name]
        }

        return acc
    }, [] as Array<string>)
